from django.apps import AppConfig


class CravensworthCoreConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'cravensworth.core'
    verbose_name = 'Django Cravensworth'
    label = 'cravensworth_core'
